#ifndef _ASM_IA64_CACHE_H
#define _ASM_IA64_CACHE_H

/*
 * Copyright (C) 1998, 1999 Hewlett-Packard Co
 * Copyright (C) 1998, 1999 David Mosberger-Tang <davidm@hpl.hp.com>
 */

/* Bytes per L1 (data) cache line.  */
#define L1_CACHE_BYTES		64

#endif /* _ASM_IA64_CACHE_H */
