#ifndef _ASM_IA64_NAMEI_H
#define _ASM_IA64_NAMEI_H

/*
 * Copyright (C) 1998, 1999 Hewlett-Packard Co
 * Copyright (C) 1998, 1999 David Mosberger-Tang <davidm@hpl.hp.com>
 */

/*
 * This dummy routine maybe changed to something useful
 * for /usr/gnemul/ emulation stuff.
 * Look at asm-sparc/namei.h for details.
 */
#define __prefix_lookup_dentry(name, lookup_flags) \
        do {} while (0)

#endif /* _ASM_IA64_NAMEI_H */
