/*
 *
 *    Copyright (c) 1999 Grant Erickson <grant@lcse.umn.edu>
 *
 *    Module name: walnut.h
 *
 *    Description:
 *	Macros, definitions, and data structures specific to the IBM PowerPC
 *      405GP "Walnut" evaluation board. Anything specific to the processor
 *      itself is defined elsewhere.
 *
 */

#ifndef	__WALNUT_H__
#define	__WALNUT_H__


#ifdef __cplusplus
extern "C" {
#endif



#ifdef __cplusplus
}
#endif

#endif /* __WALNUT_H__ */
