#ifndef _MULTIFACE_H_
#define _MULTIFACE_H_

/*
 * Defines for SerialMaster, Multiface Card II and Multiface Card III
 * The addresses given below are offsets to the board base address
 * 
 * 6.11.95 Joerg Dorchain (dorchain@mpi-sb.mpg.de)
 *
 */

#define PIA_REG_PADWIDTH 255

#define DUARTBASE 0x0000
#define PITBASE   0x0100
#define ROMBASE   0x0200
#define PIABASE   0x4000

#endif

