#ifndef _ASM_IA64_OFFSETS_H
#define _ASM_IA64_OFFSETS_H

/*
 * DO NOT MODIFY
 *
 * This file was generated by arch/ia64/tools/print_offsets.
 *
 */

#define PF_PTRACED_BIT			4

#define IA64_TASK_SIZE			3280	/* 0xcd0 */
#define IA64_PT_REGS_SIZE		400	/* 0x190 */
#define IA64_SWITCH_STACK_SIZE		560	/* 0x230 */
#define IA64_SIGINFO_SIZE		136	/* 0x88 */

#define IA64_TASK_FLAGS_OFFSET		8	/* 0x8 */
#define IA64_TASK_SIGPENDING_OFFSET	16	/* 0x10 */
#define IA64_TASK_NEED_RESCHED_OFFSET	40	/* 0x28 */
#define IA64_TASK_PROCESSOR_OFFSET	108	/* 0x6c */
#define IA64_TASK_THREAD_OFFSET		1424	/* 0x590 */
#define IA64_TASK_THREAD_KSP_OFFSET	1424	/* 0x590 */
#define IA64_TASK_PID_OFFSET		188	/* 0xbc */
#define IA64_TASK_MM_OFFSET		88	/* 0x58 */
#define IA64_PT_REGS_CR_IPSR_OFFSET	0	/* 0x0 */
#define IA64_PT_REGS_R12_OFFSET		112	/* 0x70 */
#define IA64_PT_REGS_R8_OFFSET		144	/* 0x90 */
#define IA64_PT_REGS_R16_OFFSET		176	/* 0xb0 */
#define IA64_SWITCH_STACK_B0_OFFSET	464	/* 0x1d0 */
#define IA64_SWITCH_STACK_CALLER_UNAT_OFFSET 0	/* 0x0 */
#define IA64_SIGCONTEXT_AR_BSP_OFFSET	72	/* 0x48 */
#define IA64_SIGCONTEXT_AR_RNAT_OFFSET	80	/* 0x50 */
#define IA64_SIGCONTEXT_FLAGS_OFFSET	0	/* 0x0 */
#define IA64_SIGCONTEXT_CFM_OFFSET	48	/* 0x30 */
#define IA64_SIGCONTEXT_FR6_OFFSET	560	/* 0x230 */

#endif /* _ASM_IA64_OFFSETS_H */
